#!/usr/bin/env python3

import glob
import os
import sys
from pathlib import Path

import click
from ase import Atoms
from ase.io import read
from utils.main_utils import MainUtils as mu

from delta_scf.calc_dscf import CalcDeltaSCF as cds
from delta_scf.force_occupation import Basis, ForceOccupation, Projector
from delta_scf.plot import Plot
from delta_scf.schmid_pseudo_voigt import broaden


def main(
    ctx,
    hpc,
    geometry_input,
    control_input,
    binary,
    run_location,
    spec_mol,
    constr_atom,
    spec_at_constr,
    occupation,
    n_atoms,
    basis_set,
    graph,
    print_output,
    nprocs,
    debug,
):
    """
    Point of controlling the program flow. Options used for all methods are
    also defined here.
    """

    # Pass global options to subcommands
    ctx.ensure_object(dict)

    # A geometry file must be given if specific atom indices are to be constrained
    if len(spec_at_constr) > 0:
        if not geometry_input:
            raise click.MissingParameter(
                param_hint="-e/--geometry_input", param_type="option"
            )

    # Use ASE unless both a custom geometry.in and control.in are specified
    # Also don't use ASE if a control.in is specified
    ase = True
    found_lattice_vecs = False
    if geometry_input is not None:
        found_lattice_vecs = mu.check_geom(geometry_input)
        ctx.obj["GEOM_INP"] = geometry_input.name
    else:
        ctx.obj["GEOM_INP"] = None
        ctx.obj["LATTICE_VECS"] = None

    found_k_grid = False
    if control_input is not None:
        ase = False
        found_k_grid = mu.check_control(control_input)
        ctx.obj["CONTROL_INP"] = control_input.name
    else:
        ctx.obj["CONTROL_INP"] = None

    if found_lattice_vecs or found_k_grid:
        ctx.obj["LATTICE_VECS"] = True

    # Find the structure if not given
    # Build the structure if given
    atoms = Atoms()

    if spec_mol is None and geometry_input is None:
        if "--help" not in sys.argv:
            try:
                atoms = read(f"./{run_location}/ground/geometry.in")
                print(
                    "molecule argument not provided, defaulting to using existing geometry.in"
                    " file"
                )
            except FileNotFoundError:
                raise click.MissingParameter(
                    param_hint="-m/--molecule or -e/--geometry_input",
                    param_type="option",
                )

    elif "--help" not in sys.argv and ase:
        if spec_mol is not None:
            atoms = mu.build_geometry(spec_mol)
        if geometry_input is not None:
            atoms = read(geometry_input.name)

    # Get the constrained atom element
    # TODO: support for multiple constrained atoms
    if constr_atom is None:
        for atom in atoms:
            if atom.index in spec_at_constr:
                constr_atom = atom.symbol
                break

    # Check if a binary has been specified
    if "--help" not in sys.argv:
        current_path = os.path.dirname(os.path.realpath(__file__))
        with open(f"{current_path}/aims_bin_loc.txt", "r") as f:
            try:
                bin_path = f.readlines()[0][:-1]
            except IndexError:
                bin_path = ""

        # Ensure the user has entered the path to the binary
        # If not open the user's $EDITOR to allow them to enter the path
        if not Path(bin_path).is_file() or binary or bin_path == "":
            marker = (
                "\n# Enter the path to the FHI-aims binary above this line\n"
                "# Ensure that the binary is located in the build directory of FHIaims"
            )
            bin_line = click.edit(marker)
            if bin_line is not None:
                with open(f"{current_path}/aims_bin_loc.txt", "w") as f:
                    f.write(bin_line)

                with open(f"{current_path}/aims_bin_loc.txt", "r") as f:
                    binary = f.readlines()[0]

            else:
                raise FileNotFoundError(
                    "path to the FHI-aims binary could not be found"
                )

        elif Path(bin_path).exists():
            print(f"specified binary path: {bin_path}")
            binary = bin_path

        species = f"{Path(binary).parent.parent}/species_defaults/"

        # Check if the species_defaults directory exists in the correct location
        if not Path(species).exists():
            print(
                "\nError: ensure the FHI-aims binary is in the 'build' directory of the FHI-aims"
                " source code directory, and that the 'species_defaults' directory exists"
            )
            raise NotADirectoryError(
                f"species_defaults directory not found in {Path(binary).parent.parent}"
            )

        # Create the ASE calculator
        if ase:
            aims_calc = mu.create_calc(nprocs, binary, species, basis_set)
            atoms.calc = aims_calc
            ctx.obj["CALC"] = aims_calc

        # User specified context objects
        ctx.obj["ATOMS"] = atoms
        ctx.obj["SPEC_MOL"] = spec_mol
        ctx.obj["BINARY"] = binary
        ctx.obj["RUN_LOC"] = run_location
        ctx.obj["CONSTR_ATOM"] = constr_atom
        ctx.obj["SPEC_AT_CONSTR"] = spec_at_constr
        ctx.obj["OCC"] = occupation
        ctx.obj["N_ATOMS"] = n_atoms  # TODO
        ctx.obj["BASIS_SET"] = basis_set
        ctx.obj["GRAPH"] = graph
        ctx.obj["PRINT"] = print_output
        ctx.obj["NPROCS"] = nprocs
        ctx.obj["DEBUG"] = debug
        ctx.obj["HPC"] = hpc

        # Context objects created in main()
        ctx.obj["SPECIES"] = species
        ctx.obj["ASE"] = ase


def process(
    ctx, intensity=1, asym=False, a=0.2, b=0.0, gl_ratio=0.5, omega=0.35, gmp=0.003
):
    """Calculate DSCF values and plot the simulated XPS spectra."""

    # Calculate the delta scf energies
    grenrgys = cds.read_ground(ctx.obj["RUN_LOC"])
    element, excienrgys = cds.read_atoms(
        ctx.obj["RUN_LOC"], ctx.obj["CONSTR_ATOM"], cds.contains_number
    )
    xps = cds.calc_delta_scf(element, grenrgys, excienrgys)

    if ctx.obj["RUN_LOC"] != "./":
        os.system(f"mv {element}_xps_peaks.txt {ctx.obj['RUN_LOC']}")

    if ctx.obj["GRAPH"]:
        # Apply the peak broadening
        peaks, domain = broaden(0, 1000, intensity, gl_ratio, xps, omega, asym, a, b)

        # Write out the spectrum to a text file
        # Include bin width of 0.01 eV
        data = []
        bin_val = 0.00
        for i, peak in enumerate(peaks):
            data.append(f"{str(bin_val)} {str(peak)}\n")
            bin_val += 0.01

        with open(f"{element}_xps_spectrum.txt", "w") as spec:
            spec.writelines(data)

        # Move the spectrum to the run location
        if ctx.obj["RUN_LOC"] != "./":
            os.system(f'mv {element}_xps_spectrum.txt {ctx.obj["RUN_LOC"]}/')

        print("\nplotting spectrum and calculating MABE...")
        Plot.sim_xps_spectrum(
            xps, ctx.obj["RUN_LOC"], ctx.obj["CONSTR_ATOM"], ctx.obj["AT_SPEC"][0], gmp
        )


def projector_wrapper(
    ctx, run_type, occ_type, pbc, l_vecs, spin, ks_range, control_opts
):
    """Force occupation of the Kohn-Sham states."""

    run_loc = ctx.obj["RUN_LOC"]
    geom_inp = ctx.obj["GEOM_INP"]
    control_inp = ctx.obj["CONTROL_INP"]
    atoms = ctx.obj["ATOMS"]
    found_lattice_vecs = ctx.obj["LATTICE_VECS"]
    basis_set = ctx.obj["BASIS_SET"]
    ase = ctx.obj["ASE"]
    nprocs = ctx.obj["NPROCS"]
    binary = ctx.obj["BINARY"]
    hpc = ctx.obj["HPC"]
    constr_atoms = ctx.obj["CONSTR_ATOM"]
    spec_at_constr = ctx.obj["SPEC_AT_CONSTR"]
    spec_mol = ctx.obj["SPEC_MOL"]
    species = ctx.obj["SPECIES"]
    occ = ctx.obj["OCC"]
    print_output = ctx.obj["PRINT"]

    if ase:
        calc = ctx.obj["CALC"]
    else:
        calc = None

    # Used later to redirect STDERR to /dev/null to prevent printing not converged errors
    spec_run_info = None

    # Raise a warning if no additional control options have been specified
    if len(control_opts) < 1 and control_inp is None:
        print(
            "\nWarning: no control options provided, using default options "
            "which can be found in the 'control.in' file"
        )

    # Convert control options to a dictionary
    control_opts = mu.convert_opts_to_dict(control_opts, pbc)

    # Check if the lattice vectors and k_grid have been provided
    if found_lattice_vecs or l_vecs is not None:
        if pbc is None:
            print(
                "Warning: -p/--pbc argument not given, attempting to use"
                " k_grid from control file or previous calculation"
            )

            # Try to parse the k-grid if other calculations have been run
            try:
                pbc_list = []
                for control in glob.glob(f"{run_loc}/**/control.in", recursive=True):
                    with open(control, "r") as control:
                        for line in control:
                            if "k_grid" in line:
                                pbc_list.append(line.split()[1:])

                # If different k_grids have been used for different calculations, then
                # enforce the user to provide the k_grid
                if not pbc_list.count(pbc_list[0]) == len(pbc_list):
                    raise click.MissingParameter(
                        param_hint="-p/--pbc", param_type="option"
                    )
                else:
                    pbc_list = tuple([int(i) for i in pbc_list[0]])
                    control_opts["k_grid"] = pbc_list

            except IndexError:
                raise click.MissingParameter(param_hint="-p/--pbc", param_type="option")

    if run_type == "ground":
        mu.ground_calc(
            run_loc,
            geom_inp,
            control_inp,
            atoms,
            basis_set,
            species,
            calc,
            ase,
            control_opts,
            constr_atoms,
            nprocs,
            binary,
            hpc,
            print_output,
        )

        # Ground must be run separately to hole calculations
        return

    else:  # run_type != ground
        ground_geom = f"{run_loc}/ground/geometry.in"
        ground_control = f"{run_loc}/ground/control.in"

    if len(spec_at_constr) == 0 and constr_atoms is None:
        raise click.MissingParameter(
            param_hint="-c/--constrained_atom or -s/--specific_atom_constraint",
            param_type="option",
        )

    # Convert constr_atoms to a list
    if type(constr_atoms) is not list:
        constr_atoms = [constr_atoms]

    # Create a list of element symbols to constrain
    if len(spec_at_constr) > 0:
        element_symbols = mu.get_element_symbols(ground_geom, spec_at_constr)[0]
        constr_atoms = element_symbols
    else:
        element_symbols = constr_atoms

    fo = ForceOccupation(
        element_symbols,
        run_loc,
        ground_geom,
        control_opts,
        f"{species}/defaults_2020/{basis_set}",
    )

    # Get atom indices from the ground state geometry file
    atom_specifier = fo.read_ground_inp(constr_atoms, spec_at_constr, ground_geom)

    if run_type == "init_1":
        if hpc:
            raise click.BadParameter(
                "the -h/--hpc flag is only supported for the 'hole' run type"
            )

        if len(spec_at_constr) == 0 and len(constr_atoms) == 0:
            raise click.BadParameter(
                "no atoms have been specified to constrain, please use "
                "-c/--constr_atoms or -s/--spec_at_constr options"
            )

        # Check required arguments are given in main()
        mu.check_args(("ks_range", ks_range))

        # TODO allow this for multiple constrained atoms using n_atoms
        for atom in element_symbols:
            fo.get_electronic_structure(atom)

        # Setup files required for the initialisation and hole calculations
        proj = Projector(fo)
        proj.setup_init_1(basis_set, species, ground_control)
        proj.setup_init_2(
            ks_range[0], ks_range[1], occ, occ_type, spin, found_lattice_vecs
        )
        proj.setup_hole(
            ks_range[0], ks_range[1], occ, occ_type, spin, found_lattice_vecs
        )

    spec_run_info = ""

    if run_type == "init_2":
        if hpc:
            raise click.BadParameter(
                "the -h/--hpc flag is only supported for the 'hole' run type"
            )

        if len(spec_at_constr) == 0 and len(constr_atoms) == 0:
            raise click.BadParameter(
                "no atoms have been specified to constrain, please use "
                "-c/--constr_atoms or -s/--spec_at_constr options"
            )

        for i in range(len(atom_specifier)):
            i += 1

            # Catch for if init_1 hasn't been run
            if len(glob.glob(f"{run_loc}/{constr_atoms[0]}{i}/init_1/*restart*")) < 1:
                print(
                    'init_1 restart files not found, please ensure "init_1" has been run'
                )
                raise FileNotFoundError

            if len(control_opts) > 0 or control_inp:
                # Add any additional control options to the init_2 control file
                parsed_control_opts = fo.get_control_keywords(
                    f"{run_loc}/{constr_atoms[0]}{i}/init_2/control.in"
                )
                mod_control_opts = fo.mod_keywords(control_opts, parsed_control_opts)
                control_content = fo.change_control_keywords(
                    f"{run_loc}/{constr_atoms[0]}{i}/init_2/control.in",
                    mod_control_opts,
                )

                with open(
                    f"{run_loc}/{constr_atoms[0]}{i}/init_2/control.in", "w"
                ) as control_file:
                    control_file.writelines(control_content)

            # Copy the restart files to init_2 from init_1
            os.path.isfile(
                glob.glob(f"{run_loc}/{constr_atoms[0]}{i}/init_1/*restart*")[0]
            )
            os.system(
                f"cp {run_loc}/{constr_atoms[0]}{i}/init_1/*restart* {run_loc}/{constr_atoms[0]}{i}/init_2/"
            )

        # Prevent SCF not converged errors from printing
        spec_run_info = " 2>/dev/null"

    if run_type == "hole":
        if len(spec_at_constr) == 0 and len(constr_atoms) == 0:
            raise click.BadParameter(
                "no atoms have been specified to constrain, please use the "
                "-c/--constr_atoms or -s/--spec_at_constr options"
            )

        if hpc:
            mu.check_args(("ks_range", ks_range))

            for atom in element_symbols:
                fo.get_electronic_structure(atom)

            # Setup files required for the initialisation and hole calculations
            proj = Projector(fo)
            proj.setup_init_1(basis_set, species, ground_control)
            proj.setup_init_2(ks_range[0], ks_range[1], occ, occ_type, spin, pbc)
            proj.setup_hole(ks_range[0], ks_range[1], occ, occ_type, spin, pbc)

        # Add molecule identifier to hole geometry.in
        with open(
            f"{run_loc}/{constr_atoms[0]}{atom_specifier[0]}/hole/geometry.in", "r"
        ) as hole_geom:
            lines = hole_geom.readlines()

        lines.insert(4, f"# {spec_mol}\n")

        with open(
            f"{run_loc}/{constr_atoms[0]}{atom_specifier[0]}/hole/geometry.in", "w"
        ) as hole_geom:
            hole_geom.writelines(lines)

        if hpc:
            return

        # Catch for if init_2 hasn't been run
        for i in range(len(atom_specifier)):
            i += 1
            if (
                os.path.isfile(
                    glob.glob(f"{run_loc}/{constr_atoms[0]}{i}/init_2/*restart*")[0]
                )
                is False
            ):
                print(
                    'init_2 restart files not found, please ensure "init_2" has been run'
                )
                raise FileNotFoundError

            if len(control_opts) > 0 or control_inp:
                # Add any additional control options to the hole control file
                parsed_control_opts = fo.get_control_keywords(
                    f"{run_loc}/{constr_atoms[0]}{i}/hole/control.in"
                )
                mod_control_opts = fo.mod_keywords(control_opts, parsed_control_opts)
                control_content = fo.change_control_keywords(
                    f"{run_loc}/{constr_atoms[0]}{i}/hole/control.in", mod_control_opts
                )

                with open(
                    f"{run_loc}/{constr_atoms[0]}{i}/hole/control.in", "w"
                ) as control_file:
                    control_file.writelines(control_content)

            # Copy the restart files to hole from init_2
            os.path.isfile(
                glob.glob(f"{run_loc}/{constr_atoms[0]}{i}/init_2/*restart*")[0]
            )
            os.system(
                f"cp {run_loc}/{constr_atoms[0]}{i}/init_2/*restart* {run_loc}/{constr_atoms[0]}{i}/hole/"
            )

        spec_run_info = ""

    # Run the calculation with a nice progress bar if not already run
    if (
        run_type != "ground"
        and os.path.isfile(
            f"{run_loc}/{constr_atoms[0]}{atom_specifier[0]}/{run_type}/aims.out"
        )
        is False
        and not hpc
    ):
        # Ensure that aims always runs with the following environment variables:
        os.system("export OMP_NUM_THREADS=1")
        os.system("export MKL_NUM_THREADS=1")
        os.system("export MKL_DYNAMIC=FALSE")
        os.system("ulimit -s unlimited")

        if print_output:  # Print live output of calculation
            for i in range(len(atom_specifier)):
                i += 1
                os.system(
                    f"cd {run_loc}/{constr_atoms[0]}{i}/{run_type} && mpirun -n "
                    f"{nprocs} {binary} | tee aims.out {spec_run_info}"
                )

        else:
            with click.progressbar(
                range(len(atom_specifier)), label=f"calculating {run_type}:"
            ) as prog_bar:
                for i in prog_bar:
                    i += 1
                    os.system(
                        f"cd {run_loc}/{constr_atoms[0]}{i}/{run_type} && mpirun -n "
                        f"{nprocs} {binary} > aims.out {spec_run_info}"
                    )

        # print(f"{run_type} calculations completed successfully")

    elif run_type != "ground" and not hpc:
        print(f"{run_type} calculations already completed, skipping calculation...")

    # These need to be passed to process()
    ctx.obj["RUN_TYPE"] = run_type
    ctx.obj["AT_SPEC"] = atom_specifier

    # Compute the dscf energies and plot if option provided
    if run_type == "hole":
        process(ctx)


def basis_wrapper(
    ctx,
    run_type,
    atom_index,
    occ_type,
    multiplicity,
    n_qn,
    l_qn,
    m_qn,
    ks_max,
    control_opts,
):
    """Force occupation of the basis states."""

    # It gets annoying to type the full context object out every time
    run_loc = ctx.obj["RUN_LOC"]
    geom = ctx.obj["GEOM_INP"]
    control = ctx.obj["CONTROL_INP"]
    spec_mol = ctx.obj["SPEC_MOL"]
    atoms = ctx.obj["ATOMS"]
    ase = ctx.obj["ASE"]
    species = ctx.obj["SPECIES"]
    basis_set = ctx.obj["BASIS_SET"]
    nprocs = ctx.obj["NPROCS"]
    binary = ctx.obj["BINARY"]
    hpc = ctx.obj["HPC"]
    constr_atoms = ctx.obj["CONSTR_ATOM"]
    spec_at_constr = ctx.obj["SPEC_AT_CONSTR"]
    occ = ctx.obj["OCC"]
    print_output = ctx.obj["PRINT"]

    if ase:
        calc = ctx.obj["CALC"]
    else:
        calc = None

    # Raise a warning if no additional control options have been specified
    if len(control_opts) < 1 and control is None:
        print(
            "\nWarning: no control options provided, using default options "
            "which can be found in the 'control.in' file"
        )

    # Convert control options to a dictionary
    control_opts = mu.convert_opts_to_dict(control_opts, None)

    if run_type == "ground":
        mu.ground_calc(
            run_loc,
            geom,
            control,
            atoms,
            basis_set,
            species,
            calc,
            ase,
            control_opts,
            constr_atoms,
            nprocs,
            binary,
            hpc,
        )

        # Ground must be run separately to hole calculations
        return

    else:  # run_type == 'hole'
        ground_geom = f"{run_loc}/ground/geometry.in"

    if len(spec_at_constr) == 0 and constr_atoms is None:
        raise click.MissingParameter(
            "No atoms have been specified to constrain, please provide either"
            " the -c/--constrained_atom or the -s/--specific_atom_constraint arguments",
            param_type="option",
        )

    if run_type == "hole":
        mu.check_args(
            ("atom_index", atom_index),
            ("ks_max", ks_max),
            ("n_qn", n_qn),
            ("l_qn", l_qn),
            ("m_qn", m_qn),
        )

        if len(spec_at_constr) == 0 and len(constr_atoms) == 0:
            raise click.BadParameter(
                "no atoms have been specified to constrain, please use the "
                "-c/--constr_atoms or -s/--spec_at_constr options"
            )

        # Ensure that aims always runs with the following environment variables:
        os.system("export OMP_NUM_THREADS=1")
        os.system("export MKL_NUM_THREADS=1")
        os.system("export MKL_DYNAMIC=FALSE")
        os.system("ulimit -s unlimited")

        if os.path.isfile(f"{run_loc}/ground/aims.out") is False:
            raise FileNotFoundError(
                "\nERROR: ground aims.out not found, please ensure the ground calculation has been "
                "run"
            )

        # Convert constr_atoms to a list
        if type(constr_atoms) is not list:
            constr_atoms = [constr_atoms]

        # Create a list of element symbols to constrain
        if len(spec_at_constr) > 0:
            element_symbols = mu.get_element_symbols(ground_geom, spec_at_constr)[0]
            constr_atoms = element_symbols
        else:
            element_symbols = constr_atoms

        # Create the directories required for the hole calculation
        fo = ForceOccupation(
            element_symbols,
            run_loc,
            ground_geom,
            control_opts,
            f"{species}/defaults_2020/{basis_set}",
        )

        # Get atom indices from the ground state geometry file
        atom_specifier = fo.read_ground_inp(constr_atoms, spec_at_constr, ground_geom)

        if (
            os.path.isfile(f"{run_loc}/{constr_atoms[0]}{atom_specifier[0]}/aims.out")
            is True
        ):
            print("hole calculations already completed, skipping calculation...")

        basis = Basis(fo)
        basis.setup_basis(multiplicity, n_qn, l_qn, m_qn, occ, ks_max, occ_type)

        # Add molecule identifier to hole geometry.in
        with open(
            f"{run_loc}{constr_atoms[0]}{atom_specifier[0]}/geometry.in", "r"
        ) as hole_geom:
            lines = hole_geom.readlines()

        lines.insert(4, f"# {spec_mol}\n")

        with open(
            f"{run_loc}/{constr_atoms[0]}{atom_specifier[0]}/geometry.in", "w"
        ) as hole_geom:
            hole_geom.writelines(lines)

        # TODO allow multiple constraints using n_atoms

        for i in range(len(atom_specifier)):
            i += 1

            if len(control_opts) > 0 or control:
                # Add any additional control options to the hole control file
                parsed_control_opts = fo.get_control_keywords(
                    f"{run_loc}/{constr_atoms[0]}{i}/control.in"
                )
                mod_control_opts = fo.mod_keywords(control_opts, parsed_control_opts)
                control_content = fo.change_control_keywords(
                    f"{run_loc}/{constr_atoms[0]}{i}/control.in", mod_control_opts
                )

                with open(
                    f"{run_loc}/{constr_atoms[0]}{i}/control.in", "w"
                ) as control_file:
                    control_file.writelines(control_content)

        if not hpc:  # Run the hole calculation
            if print_output:  # Print live output of calculation
                for i in range(len(atom_specifier)):
                    i += 1
                    os.system(
                        f"cd {run_loc}/{constr_atoms[0]}{i} && mpirun -n "
                        f"{nprocs} {binary} | tee aims.out"
                    )

            else:
                with click.progressbar(
                    range(len(atom_specifier)), label="calculating basis hole:"
                ) as prog_bar:
                    for i in prog_bar:
                        i += 1
                        os.system(
                            f"cd {run_loc}/{constr_atoms[0]}{i} && mpirun -n "
                            f"{nprocs} {binary} > aims.out"
                        )

        # These need to be passed to process()
        ctx.obj["RUN_TYPE"] = run_type
        ctx.obj["AT_SPEC"] = atom_specifier

    # Compute the dscf energies and plot if option provided
    if run_type == "hole":
        process(ctx)



import os


class CalcDeltaSCF:
    """Calculate delta-scf energies from the output of ground and excited state calculations"""

    @staticmethod
    def read_ground(calc_path):
        """Get the ground state energy."""

        with open(f"{calc_path}ground/aims.out", "r", encoding="utf-8") as ground:
            lines = ground.readlines()

        grenrgys = None

        for line in lines:
            # Get the energy
            if "s.c.f. calculation      :" in line:
                grenrgys = float(line.split()[-2])

        if grenrgys is None:
            raise ValueError("No ground state energy found.")

        print("Ground state calculated energy (eV):")
        print(round(grenrgys, 3))
        print()

        return grenrgys

    @staticmethod
    def contains_number(string):
        """Check if a number is in a string."""

        for character in string:
            if character.isdigit():
                return True

    @staticmethod
    def read_atoms(calc_path, element, contains_number):
        """Get the excited state energies."""
        dir_list = os.listdir(calc_path)
        energy = "s.c.f. calculation      :"
        excienrgys = []

        # Read each core hole dir
        for directory in dir_list:
            if element in directory and contains_number(directory) is True:
                # Try reading output file from basis, then projector file structure
                if os.path.exists(f"{calc_path}{directory}/aims.out"):
                    with open(
                        f"{calc_path}{directory}/aims.out", "r", errors="ignore"
                    ) as out:
                        lines = out.readlines()
                else:
                    with open(
                        f"{calc_path}{directory}/hole/aims.out", "r", errors="ignore"
                    ) as out:
                        lines = out.readlines()

                for line in lines:
                    # Get the energy
                    if energy in line:
                        excienrgys.append(float(line.split()[-2]))

        # Remove duplicate binding energies from list
        excienrgys = list(dict.fromkeys(excienrgys))

        print("Core hole calculated energies (eV):", *excienrgys, sep="\n")

        return element, excienrgys

    @staticmethod
    def calc_delta_scf(element, grenrgys, excienrgys):
        """Calculate delta scf and write to a file."""

        xps = []

        for i in excienrgys:
            xps.append(i - grenrgys)

        print("\nDelta-SCF energies (eV):")

        for i, be in enumerate(xps):
            xps[i] = str(round(be, 3))
            print(xps[i])

        with open(element + "_xps_peaks.txt", "w") as file:
            file.writelines(xps)

        return [float(be) for be in xps]



"""Automate creation of files for FOP calculations in FHI-aims."""

import glob
import os
import shutil
import subprocess

import numpy as np
import yaml


class ForceOccupation:
    """Manipulate FHIaims input files to setup basis and projector calculations."""

    def __init__(self, element_symbols, run_loc, geometry, ad_cont_opts, species):
        self.element_symbols = element_symbols
        self.run_loc = run_loc
        self.geometry = geometry
        self.ad_cont_opts = ad_cont_opts
        self.species = species

        # Convert k_grid key to a string from a tuple
        # Writing the options for a hole calculation doesn't use ASE, so it must be
        # converted to a string here
        if "k_grid" in ad_cont_opts.keys():
            ad_cont_opts["k_grid"] = " ".join(map(str, ad_cont_opts["k_grid"]))

        self.new_control = f"{self.run_loc}/ground/control.in.new"
        self.atom_specifier = []

        # Find the root directory of the package
        self.current_path = os.path.dirname(os.path.realpath(__file__))

        # All supported elements
        with open(f"{self.current_path}/elements.yml", "r") as elements:
            self.elements = yaml.load(elements, Loader=yaml.SafeLoader)

    def read_ground_inp(self, constr_atoms, spec_at_constr, geometry_path):
        """Find the number of atoms in the geometry file."""

        # For if the user supplied element symbols to constrain
        if constr_atoms is not None:
            # Check validity of specified elements
            for atom in constr_atoms:
                if atom not in self.elements:
                    raise ValueError("invalid element specified")

            print(
                "atoms argument not specified, defaulting to all target atoms in geometry.in"
            )

            # Constrain all atoms of the target element
            for atom in constr_atoms:
                with open(geometry_path, "r") as geom_in:
                    atom_counter = 0

                    for line in geom_in:
                        spl = line.split()

                        if len(spl) > 0 and "atom" in spl[0]:
                            atom_counter += 1
                            element = spl[-1]  # Identify atom
                            identifier = spl[0]  # Extra check that line is an atom

                            if "atom" in identifier and element == atom:
                                self.atom_specifier.append(atom_counter)

        # For if the user supplied atom indices to constrain
        if len(spec_at_constr) > 0:
            # Check validity of specified elements
            for atom in self.element_symbols:
                if atom not in self.elements:
                    raise ValueError("invalid element specified")

            self.atom_specifier = list(spec_at_constr)

        print("specified atom indices:", self.atom_specifier)

        return self.atom_specifier

    def get_electronic_structure(self, atom):
        """Get valence electronic structure of target atom."""
        # Adapted from scipython.com question P2.5.12

        # Get the atomic number
        self.atom_index = self.elements.index(str(atom)) + 1

        # Letters identifying subshells
        l_letter = ["s", "p", "d", "f", "g"]

        # Create and order a list of tuples, (n+l, n, l), corresponding to the order
        # in which the corresponding orbitals are filled using the Madelung rule.
        nl_pairs = []

        for n in range(1, 8):
            for l in range(n):
                nl_pairs.append((n + l, n, l))

        nl_pairs.sort()

        # inl indexes the subshell in the nl_pairs list; nelec is the number of
        # electrons currently inhabiting this subshell
        inl, nelec = 0, 0

        # start with the 1s orbital
        n, l = 1, 0

        # a list of orbitals and the electrons they contain for a configuration
        config = [["1s", 0]]

        # Store the most recent Noble gas configuration encountered in a tuple with the
        # corresponding element symbol
        noble_gas_config = ("", "")
        s_config = ""

        for i in range(len(self.elements[: self.atom_index])):
            nelec += 1

            # this subshell is now full
            if nelec > 2 * (2 * l + 1):
                # The most recent configuration was for a Noble gas: store it
                if l == 1:
                    # Turn a list of orbital, nelec pairs into a configuration string
                    noble_gas_config = (
                        ".".join(["{:2s}{:d}".format(*e) for e in config]),
                        "[{}]".format(self.elements[i - 1]),
                    )

                # Start a new subshell
                inl += 1
                _, n, l = nl_pairs[inl]
                config.append(["{}{}".format(n, l_letter[l]), 1])
                nelec = 1

            # add an electron to the current subshell
            else:
                config[-1][1] += 1

            # Turn config into a string
            s_config = ".".join(["{:2s}{:d}".format(*e) for e in config])

            # Replace the heaviest Noble gas configuration with its symbol
            s_config = s_config.replace(*noble_gas_config)

        output = list(s_config.split(".").pop(-1))

        # Correct transition metal atoms where the s orbital is filled before d
        if output[1] == "d":
            if output[2] == "4" or output[2] == "9":
                output[0] = str(int(output[0]) + 1)
                output[1] = "s"
                output[2] = "1"

        self.valence = f"    valence      {output[0]}  {output[1]}   {output[2]}.1\n"

    @staticmethod
    def add_additional_basis(current_path, elements, content, target_atom):
        """Add an additional basis set for the core hole calculation."""

        # Check the additional functions haven't already been added to control
        for line in content:
            if "# Additional basis functions for atom with a core hole" in line:
                return

        # Get the additional basis set
        if "utils" in current_path.split("/"):
            with open(f"{current_path}/../delta_scf/add_basis_functions.yml", "r") as f:
                ad_basis = yaml.safe_load(f)
        else:
            with open(f"{current_path}/add_basis_functions.yml", "r") as f:
                ad_basis = yaml.safe_load(f)

        if [*target_atom][-1][0] == "1":
            root_target_atom = "".join([*target_atom][:-1])
        else:
            root_target_atom = target_atom

        try:
            el_ad_basis = ad_basis[root_target_atom]
        except KeyError:
            print(
                f"Warning: the additional basis set for {target_atom} is not yet supported."
                " The calculation will continue without the additional core-hole basis set."
            )
            el_ad_basis = ""
            pass

        # Add the additional basis set before the 5th line of '#'s after the species
        # and element line defining the start of the basis set. This is the only
        # consistent marker across all the basis sets after which the additional basis
        # set can be added.
        basis_def_start = 0
        for i, line in enumerate(content):
            if "species" in line and target_atom in line:
                basis_def_start = i
                break

        # Get the atomic number of the target atom
        atom_index = elements.index(str(root_target_atom)) + 1

        # prefix 0 if atom_index is less than 10
        if atom_index < 10:
            atom_index = f"0{atom_index}"

        # Append a separator to the end of the file
        # This helps with adding the additional basis set in the correct positions for
        # some basis sets.
        separator = (
            "#######################################################################"
            "#########"
        )
        content.append(separator + "\n")

        # Find the line which contains the appropriate row of '#'s after the species and element
        div_counter = 0
        insert_point = 0
        for i, line in enumerate(content[basis_def_start:]):
            if separator in line:
                if "species" in line and target_atom in line:
                    break

                if div_counter == 3:
                    insert_point = i + basis_def_start

                if div_counter < 4:
                    div_counter += 1

                else:
                    insert_point = i + basis_def_start
                    break

        # Append the additional basis set to the end of the basis set in the control file
        if (
            el_ad_basis != ""
            and basis_def_start != 0
            and insert_point != 0
            and div_counter != 0
        ):
            content.insert(insert_point, f"{el_ad_basis}\n")
            return content

        else:
            print(
                "Warning: there was an error with adding the additional basis functions"
            )
            return content

    @staticmethod
    def get_control_keywords(control):
        """Get the keywords in a control.in file"""

        # Find and replace keywords in control file
        with open(control, "r") as read_control:
            content = read_control.readlines()

        # Get keywords
        opts = {}
        for line in content:
            spl = line.split()

            # Break when basis set definitions start
            if (
                "############################################################"
                "####################" in line
            ):
                break

            # Add the dictionary value as a string
            if len(spl) > 1 and "#" not in spl[0]:
                if len(spl[1:]) > 1:
                    opts[spl[0]] = " ".join(spl[1:])

                else:
                    opts[spl[0]] = spl[1]

        return opts

    @staticmethod
    def mod_keywords(ad_cont_opts, opts):
        """Allow users to modify and add keywords"""

        for key in list(ad_cont_opts.keys()):
            opts.update({key: ad_cont_opts[key]})

        return opts

    @staticmethod
    def change_control_keywords(control, opts):
        """Modify the keywords in a control.in file from a dictionary of options."""

        # Find and replace keywords in control file
        with open(control, "r") as read_control:
            content = read_control.readlines()

        divider_1 = "#==============================================================================="
        divider_2 = "################################################################################"

        # Change keyword lines
        for opt in opts:
            ident = 0

            for j, line in enumerate(content):
                spl = line.split()

                # print(opts)
                # print(opts[opt])
                # print(opt)

                if opt in spl:
                    content[j] = f"{opt:<34} {opts[opt]}\n"
                    break

                # Marker if ASE was used so keywords will be added in the same
                # place as the others
                elif divider_1 in line:
                    ident += 1
                    if ident == 3:
                        content.insert(j, f"{opt:<34} {opts[opt]}\n")
                        break

                # Marker for non-ASE input files so keywords will be added
                # before the basis set definitions
                elif divider_2 in line:
                    ident += 1
                    content.insert(j - 1, f"{opt:<34} {opts[opt]}\n")
                    break

            # Ensure keywords are added in all other instances
            # if ident == 0:
            #     content.insert(0, f"{opt:<34} {opts[opt]}\n")

        return content

    @staticmethod
    def add_partial_charge(content, target_atom, at_num, atom_valence, partial_charge):
        """Add a partial charge to a basis set in a control.in file."""

        # Ensure returned variables are bound
        nuclear_index = 0
        valence_index = 0
        nucleus = ""

        for j, line in enumerate(content):
            spl = line.split()

            if target_atom + "1" in spl:
                # Add to nucleus
                if f"    nucleus             {at_num}\n" in content[j:]:
                    nuclear_index = (
                        content[j:].index(f"    nucleus             {at_num}\n") + j
                    )
                    nucleus = content[nuclear_index]  # save for hole
                    content[
                        nuclear_index
                    ] = f"    nucleus             {at_num + partial_charge}\n"
                elif f"    nucleus      {at_num}\n" in content[j:]:
                    nuclear_index = (
                        content[j:].index(f"    nucleus      {at_num}\n") + j
                    )
                    nucleus = content[nuclear_index]  # save for hole
                    content[
                        nuclear_index
                    ] = f"    nucleus      {at_num + partial_charge}\n"

                # Add to valence orbital
                if "#     ion occupancy\n" in content[j:]:
                    vbs_index = content[j:].index("#     valence basis states\n") + j
                    io_index = content[j:].index("#     ion occupancy\n") + j

                    # Check which orbital to add 0.1 to
                    principle_qns = np.array([])
                    azimuthal_orbs = np.array([])
                    azimuthal_qns = np.zeros(io_index - vbs_index - 1)
                    azimuthal_refs = {"s": 1, "p": 2, "d": 3, "f": 4, "g": 5}

                    # Get azimuthal and principle quantum numbers
                    for count, valence_orbital in enumerate(
                        content[vbs_index + 1 : io_index]
                    ):
                        principle_qns = np.append(
                            principle_qns,
                            np.array(valence_orbital.split()[1]),
                        ).astype(int)
                        azimuthal_orbs = np.append(
                            azimuthal_orbs,
                            np.array(valence_orbital.split()[2]),
                        )
                        azimuthal_qns[count] = azimuthal_refs[azimuthal_orbs[count]]
                        azimuthal_qns = azimuthal_qns.astype(int)

                    # Find the orbital with highest principle and azimuthal qn
                    highest_n = np.amax(principle_qns)
                    highest_nuclear_index = np.where(principle_qns == highest_n)

                    # Check for highest l if 2 orbitals have the same n
                    if len(highest_nuclear_index[0]) > 1:
                        highest_l = np.amax(azimuthal_qns)
                        highest_l_index = np.where(azimuthal_qns == highest_l)
                        addition_state = np.intersect1d(
                            highest_nuclear_index, highest_l_index
                        )[0]
                    else:
                        addition_state = highest_nuclear_index[0][0]

                    # Add the 0.1 electron
                    valence_index = vbs_index + addition_state + 1
                    content[valence_index] = atom_valence
                    break

        return nuclear_index, valence_index, nucleus, content


class Projector(ForceOccupation):
    """Create input files for projector calculations."""

    def __init__(self, parent_instance):
        """Inherit all the variables from an instance of the parent class"""
        vars(self).update(vars(parent_instance))

    def setup_init_1(self, basis_set, defaults, control):
        """Write new directories and control files for the first initialisation calculation."""

        # Default control file options
        opts = {
            # "xc": "pbe",
            # "spin": "collinear",
            # "default_initial_moment": 0,
            "charge": 0.1,
            "sc_iter_limit": 500,
            "sc_init_iter": 75,
            "restart_write_only": "restart_file",
            "restart_save_iterations": 5,
        }

        # Add or change user-specified keywords to the control file
        opts = self.mod_keywords(self.ad_cont_opts, opts)

        # Create a new intermediate file and write basis sets to it
        shutil.copyfile(control, self.new_control)

        # Loop over each element to constrain
        for el in self.element_symbols:
            # Find species defaults location from location of binary
            basis_set = glob.glob(f"{defaults}/defaults_2020/{basis_set}/*{el}_default")
            bash_add_basis = f"cat {basis_set[0]}"

            # Create a new intermediate control file
            new_control = open(self.new_control, "a")
            subprocess.run(bash_add_basis.split(), check=True, stdout=new_control)

            # Change basis set label for core hole atom
            with open(self.new_control, "r") as read_control:
                new_basis_content = read_control.readlines()

            found_target_atom = False

            for j, line in enumerate(new_basis_content):
                spl = line.split()

                if len(spl) > 1:
                    if "species" == spl[0] and el == spl[1]:
                        if found_target_atom is False:
                            new_basis_content[j] = f"  species        {el}1\n"
                            found_target_atom = True

            # Write it to intermediate control file
            with open(self.new_control, "w") as write_control:
                write_control.writelines(new_basis_content)

            # Loop over each individual atom to constrain
            for i in self.atom_specifier:
                # TODO: fix this for individual atom constraints

                i1_control = f"{self.run_loc}/{el}{i}/init_1/control.in"
                i1_geometry = f"{self.run_loc}/{el}{i}/init_1/geometry.in"

                # Create new directory and control file for init_1 calc
                os.makedirs(f"{self.run_loc}/{el}{i}/init_1", exist_ok=True)
                shutil.copyfile(
                    self.new_control,
                    i1_control,
                )
                # Create new geometry file for init_1 calc
                shutil.copyfile(f"{self.run_loc}/ground/geometry.in", i1_geometry)

                # Change geometry file
                with open(i1_geometry, "r") as read_geom:
                    geom_content = read_geom.readlines()

                # Change core hole atom to {atom}{num}
                atom_counter = 0
                for j, line in enumerate(geom_content):
                    spl = line.split()

                    if "atom" in line and el in line:
                        if atom_counter + 1 == i:
                            partial_hole_atom = f" {el}1\n"
                            geom_content[j] = " ".join(spl[0:-1]) + partial_hole_atom

                        atom_counter += 1

                with open(i1_geometry, "w") as write_geom:
                    write_geom.writelines(geom_content)

                # Change control file
                control_content = self.change_control_keywords(i1_control, opts)

                # Add additional core-hole basis functions
                control_content = self.add_additional_basis(
                    self.current_path, self.elements, control_content, f"{el}1"
                )

                (
                    self.n_index,
                    self.v_index,
                    self.nucleus,
                    control_content,
                ) = self.add_partial_charge(
                    control_content,
                    el,
                    self.elements.index(el) + 1,
                    self.valence,
                    opts["charge"],
                )

                with open(i1_control, "w") as write_control:
                    write_control.writelines(control_content)

        print("init_1 files written successfully")

    def setup_init_2(self, ks_start, ks_stop, occ, occ_type, spin, pbc):
        """Write new directories and control files for the second initialisation calculation."""

        ks_method = ""
        if occ_type == "force_occupation_projector":
            ks_method = "serial"
        if occ_type == "deltascf_projector" and pbc is False:
            ks_method = "parallel"
        if occ_type == "deltascf_projector" and pbc is True:
            ks_method = "serial"

        # Loop over each element to constrain
        for el in self.element_symbols:
            # Loop over each individual atom to constrain
            for i in self.atom_specifier:
                opts = {
                    # "xc": "pbe",
                    # "spin": "collinear",
                    # "default_initial_moment": 0,
                    "charge": 1.1,
                    "sc_iter_limit": 1,
                    occ_type: f"{ks_start} {spin} {occ} {ks_start} {ks_stop}",
                    "KS_method": ks_method,
                    "restart": "restart_file",
                    "restart_save_iterations": 1,
                }

                # Add or change user-specified keywords to the control file
                opts = self.mod_keywords(self.ad_cont_opts, opts)

                i2_control = f"{self.run_loc}/{el}{i}/init_2/control.in"

                # Create new directory for init_2 calc
                os.makedirs(f"{self.run_loc}/{el}{i}/init_2", exist_ok=True)
                shutil.copyfile(self.new_control, i2_control)
                shutil.copyfile(
                    f"{self.run_loc}/{el}{i}/init_1/geometry.in",
                    f"{self.run_loc}/{el}{i}/init_2/geometry.in",
                )

                # Change control file
                control_content = self.change_control_keywords(i2_control, opts)

                # Add additional core-hole basis functions
                control_content = self.add_additional_basis(
                    self.current_path, self.elements, control_content, f"{el}1"
                )

                # Add partial charge to the control file
                _, _, _, control_content = self.add_partial_charge(
                    control_content,
                    el,
                    self.elements.index(el) + 1,
                    self.valence,
                    opts["charge"] - 1,
                )

                with open(i2_control, "w") as write_control:
                    write_control.writelines(control_content)

        print("init_2 files written successfully")

    def setup_hole(self, ks_start, ks_stop, occ, occ_type, spin, pbc):
        """Write new hole directories and control files for the hole calculation."""

        # Calculate original valence state
        val_spl = self.valence.split(".")
        del val_spl[-1]
        val_spl.append(".\n")
        self.valence = "".join(val_spl)

        # The new basis method should utilise ks method parallel
        ks_method = ""
        if occ_type == "force_occupation_projector":
            ks_method = "serial"
        if occ_type == "deltascf_projector" and pbc is False:
            ks_method = "parallel"
        if occ_type == "deltascf_projector" and pbc is True:
            ks_method = "serial"

        # Loop over each element to constrain
        for el in self.element_symbols:
            # Loop over each individual atom to constrain
            for i in self.atom_specifier:
                opts = {
                    # "xc": "pbe",
                    # "spin": "collinear",
                    # "default_initial_moment": 0,
                    "charge": 1.0,
                    "sc_iter_limit": 500,
                    "sc_init_iter": 75,
                    occ_type: f"{ks_start} {spin} {occ} {ks_start} {ks_stop}",
                    "KS_method": ks_method,
                    "restart_read_only": "restart_file",
                    "output": "cube spin_density",
                }

                # Add or change user-specified keywords to the control file
                opts = self.mod_keywords(self.ad_cont_opts, opts)

                # Location of the hole control file
                h_control = f"{self.run_loc}/{el}{i}/hole/control.in"

                # Create new directory for hole calc
                os.makedirs(f"{self.run_loc}/{el}{i}/hole", exist_ok=True)
                shutil.copyfile(
                    f"{self.run_loc}/{el}{i}/init_1/geometry.in",
                    f"{self.run_loc}/{el}{i}/hole/geometry.in",
                )
                shutil.copyfile(self.new_control, h_control)

                with open(h_control, "r") as read_control:
                    control_content = read_control.readlines()

                # Set nuclear and valence orbitals back to integer values
                control_content[self.n_index] = self.nucleus
                control_content[self.v_index] = self.valence

                # Change control file
                control_content = self.change_control_keywords(h_control, opts)

                # Add additional core-hole basis functions
                control_content = self.add_additional_basis(
                    self.current_path, self.elements, control_content, f"{el}1"
                )

                # Remove partial charge from the control file
                _, _, _, control_content = self.add_partial_charge(
                    control_content, el, self.elements.index(el) + 1, self.valence, 0
                )

                # Write the data to the file
                with open(h_control, "w") as write_control:
                    write_control.writelines(control_content)

        print("hole files written successfully")


class Basis(ForceOccupation):
    """Create input files for basis calculations."""

    def __init__(self, parent_instance):
        """Inherit all the variables from an instance of the parent class"""
        vars(self).update(vars(parent_instance))

    def setup_basis(self, spin, n_qn, l_qn, m_qn, occ_no, ks_max, occ_type):
        """Write new directories and control files for basis calculations."""

        # The new basis method should utilise ks method parallel
        ks_method = ""
        if occ_type == "force_occupation_basis":
            ks_method = "serial"
        if occ_type == "deltascf_basis":
            ks_method = "parallel"

        # Iterate over each constrained atom
        for el in self.element_symbols:
            # Loop over each individual atom to constrain
            for i in range(len(self.atom_specifier)):
                # Default control file options
                opts = {
                    # "xc": "pbe",
                    # "spin": "collinear",
                    # "default_initial_moment": 0,
                    "charge": 1.0,
                    "sc_iter_limit": 500,
                    "sc_init_iter": 75,
                    occ_type: f"{self.atom_specifier[i]} {spin} atomic {n_qn} {l_qn} {m_qn} {occ_no} {ks_max}",
                    "KS_method": ks_method,
                }

                # Allow users to modify and add keywords
                opts = self.mod_keywords(self.ad_cont_opts, opts)

                i += 1

                control = f"{self.run_loc}/{el}{i}/control.in"

                # Create new directories and .in files for each constrained atom
                os.makedirs(f"{self.run_loc}/{el}{i}/", exist_ok=True)
                shutil.copyfile(
                    f"{self.run_loc}/ground/control.in",
                    control,
                )
                shutil.copyfile(
                    f"{self.run_loc}/ground/geometry.in",
                    f"{self.run_loc}/{el}{i}/geometry.in",
                )

                # Change control file
                content = self.change_control_keywords(control, opts)

                # Write the data to the file
                with open(control, "w") as write_control:
                    write_control.writelines(content)

        print("files and directories written successfully")



#!/usr/bin/env python3

import click
from utils.custom_click import MutuallyExclusive as me

from delta_scf.aims_dscf import basis_wrapper, main, process, projector_wrapper


@click.group()
@click.option(
    "-h",
    "--hpc",
    cls=me,
    mutually_exclusive=["--binary"],
    is_flag=True,
    help="setup a calculation primarily for use on a HPC cluster WITHOUT "
    "running the calculation",
)
@click.option(
    "-m",
    "--molecule",
    "spec_mol",
    cls=me,
    mutually_exclusive=["--geometry_input"],
    type=str,
    help="molecule to be used in the calculation",
)
@click.option(
    "-e",
    "--geometry_input",
    cls=me,
    mutually_exclusive=["--molecule"],
    nargs=1,
    type=click.File(),
    help="specify a custom geometry.in instead of using a structure from PubChem or ASE",
)
@click.option(
    "-i",
    "--control_input",
    nargs=1,
    type=click.File(),
    help="specify a custom control.in instead of automatically generating one",
)
@click.option(
    "-y",
    "--binary",
    cls=me,
    mutually_exclusive=["hpc"],
    is_flag=True,
    help="modify the path to the FHI-aims binary",
)
@click.option(
    "-r",
    "--run_location",
    default="./",
    show_default=True,
    type=click.Path(file_okay=False, dir_okay=True),
    help="Optionally specify a custom location to run the calculation",
)
@click.option(
    "-c",
    "--constrained_atom",
    "constr_atom",
    cls=me,
    mutually_exclusive=["spec_at_constr"],
    type=str,
    # multiple=True,  # TODO: allow for multiple atoms to be constrained
    help="atom to constrain; constrain all atoms of this element",
)
@click.option(
    "-s",
    "--specific_atom_constraint",
    "spec_at_constr",
    cls=me,
    mutually_exclusive=["constr_atom"],
    multiple=True,
    type=click.IntRange(min=1, max_open=True),
    help="specify specific atoms to constrain by their index in a geometry file",
)
@click.option(
    "-o",
    "--occupation",
    type=float,
    default=0.0,
    show_default=True,
    help="occupation value of the core hole",
)
@click.option(
    "-a",
    "--n_atoms",
    type=click.IntRange(1),
    default=1,
    show_default=True,
    help="number of atoms to constrain per calculation",
)
@click.option(
    "-b",
    "--basis_set",
    default="tight",
    show_default=True,
    type=click.Choice(["light", "intermediate", "tight", "really_tight"]),
    help="the basis set to use for the calculation",
)
@click.option("-g", "--graph", is_flag=True, help="print out the simulated XPS spectra")
@click.option(
    "-p",
    "--print_output",
    is_flag=True,
    help="print the live output of the calculation",
    # TODO
)
@click.option(
    "-n",
    "--nprocs",
    default=4,
    show_default=True,
    type=int,
    help="number of processors to use",
)
@click.option(
    "-d", "--debug", is_flag=True, help="for developer use: print debug information"
)
@click.pass_context
def cli(
    ctx,
    hpc,
    geometry_input,
    control_input,
    binary,
    run_location,
    spec_mol,
    constr_atom,
    spec_at_constr,
    occupation,
    n_atoms,
    basis_set,
    graph,
    print_output,
    nprocs,
    debug,
):
    """An interface to automate core-hole constrained occupation methods in
    FHI-aims.

    There is functionality to use both the older and soon-to-be deprecated
    forced_occupation_basis and forced_occupation_projector methods, as well as
    the newer and faster deltascf_basis and deltascf_projector methods. This was
    originally written as a testing application and has since developed into an
    application for automating the basis and projector methods.

    Functionality has been included to run on both a local machine, or a HPC
    cluster. Installation is automated using Poetry, and structures from the ASE
    and PubChem databases can be used to generate the geometry.in file, or the
    geometry.in and control.in can be manually created and passed to this program.
    For full documentation, please refer to the README.md.

    Copyright \u00A9 2022-2023, Dylan Morgan dylan.morgan@warwick.ac.uk
    """

    main(
        ctx,
        hpc,
        geometry_input,
        control_input,
        binary,
        run_location,
        spec_mol,
        constr_atom,
        spec_at_constr,
        occupation,
        n_atoms,
        basis_set,
        graph,
        print_output,
        nprocs,
        debug,
    )


@cli.command()
@click.option(
    "-r",
    "--run_type",
    required=True,
    type=click.Choice(["ground", "init_1", "init_2", "hole"]),
    help="the type of calculation to perform",
)
@click.option(
    "-t",
    "--occ_type",
    default="deltascf_projector",
    show_default=True,
    type=click.Choice(["deltascf_projector", "force_occupation_projector"]),
    help="select whether the old or new occupation routine is used",
)
@click.option(
    "-p",
    "--pbc",
    nargs=3,
    type=int,
    help="give the k-grid for a periodic calculation",
)
@click.option(
    "-l",
    "--lattice_vectors",
    "l_vecs",
    nargs=1,
    type=list,
    help="provide the lattice vectors in a 3x3 matrix",
)
@click.option(
    "-s",
    "--spin",
    type=click.Choice(["1", "2"]),
    default="1",
    show_default=True,
    help="set the spin channel of the constraint",
)
@click.option(
    "-k",
    "--ks_range",
    nargs=2,
    type=click.IntRange(1),
    help="range of Kohn-Sham states to constrain taken with 2 arguments",
)
@click.option(
    "-c",
    "--control_opts",
    multiple=True,
    type=str,
    help="provide additional options to be used in 'control.in' in a key=value format",
)
@click.pass_context
def projector(ctx, run_type, occ_type, pbc, l_vecs, spin, ks_range, control_opts):
    """Directly force occupation of the Kohn-Sham states."""

    projector_wrapper(
        ctx, run_type, occ_type, pbc, l_vecs, spin, ks_range, control_opts
    )


@cli.command()
@click.option(
    "-r",
    "--run_type",
    required=True,
    type=click.Choice(["ground", "hole"]),
    help="select the type of calculation to perform",
)
@click.option(
    "-a", "--atom_index", type=int, help="index of the (first) atom to constrain"
)
@click.option(
    "-o",
    "--occ_type",
    default="deltascf_basis",
    show_default=True,
    type=click.Choice(["deltascf_basis", "force_occupation_basis"]),
    help="select whether the old or new occupation routine is used",
)
@click.option(
    "-s",
    "--spin",
    type=click.Choice(["1", "2"]),
    default="1",
    show_default=True,
    help="set the spin channel for the constraint",
)
@click.option(
    "-n",
    "--n_quantum_number",
    "n_qn",
    type=int,
    help="principal quantum number of constrained state",
)
@click.option(
    "-l",
    "--l_quantum_number",
    "l_qn",
    type=int,
    help="orbital momentum quantum number of constrained state",
)
@click.option(
    "-m",
    "--m_quantum_number",
    "m_qn",
    type=int,
    help="magnetic quantum number for projection of orbital momentum",
)
@click.option(
    "-k",
    "--ks_max",
    type=click.IntRange(1),
    help="maximum Kohn-Sham state to constrain",
)
@click.option(
    "-c",
    "--control_opts",
    multiple=True,
    type=str,
    help="provide additional options to be used in 'control.in'",
)
@click.pass_context
def basis(
    ctx,
    run_type,
    atom_index,
    occ_type,
    spin,
    n_qn,
    l_qn,
    m_qn,
    ks_max,
    control_opts,
):
    """Force occupation of Kohn-Sham states through basis functions."""

    basis_wrapper(
        ctx,
        run_type,
        atom_index,
        occ_type,
        spin,
        n_qn,
        l_qn,
        m_qn,
        ks_max,
        control_opts,
    )


@cli.command()
@click.option(
    "-A",
    "--intensity",
    default=1,
    type=float,
    show_default=True,
    help="Set as 1 for all subpeaks if they are non-degenerate",
)
@click.option(
    "-s",
    "--asym",
    is_flag=True,
    default=False,
    show_default=True,
    help="Simulate the XPS spectrum with asymmetry",
)
@click.option(
    "-a",
    "--asym_param",
    "a",
    default=0.2,
    type=float,
    show_default=True,
    help="Define the asymmetry parameter",
)
@click.option(
    "-b",
    "--asym_trans_param",
    "b",
    default=0.0,
    type=float,
    show_default=True,
    help="Define the asymmetry translation parameter",
)
@click.option(
    "-m",
    "--gl_ratio",
    default=0.5,
    type=click.FloatRange(min=0, max=1, clamp=True),
    show_default=True,
    help="Set the mixing parameter for the Gaussian-Lorentzian functions",
)
@click.option(
    "-o",
    "--omega",
    default=0.35,
    type=click.FloatRange(min=0, max=1, clamp=True),
    show_default=True,
    help="Full width at half maximum value",
)
@click.option(
    "-g",
    "--gmp",
    default=0.003,
    type=click.FloatRange(min=0, max_open=True),
    show_default=True,
    help="Global minimum percentage",
)
@click.pass_context
def plot(ctx, intensity, asym, a, b, gl_ratio, omega, gmp):
    """Plot the simulated XPS spectra."""

    process(ctx, intensity, asym, a, b, gl_ratio, omega, gmp)


if __name__ == "__main__":
    cli()



import math

import matplotlib.pyplot as plt
import numpy as np


class Plot:
    """Plot the XPS spectrum"""

    @staticmethod
    def sim_xps_spectrum(xps, run_loc, targ_at, at_spec, gmp):
        x_axis_aims = np.loadtxt(f"{run_loc}/{targ_at}_xps_spectrum.txt", usecols=(0))
        y_axis_aims = np.loadtxt(f"{run_loc}/{targ_at}_xps_spectrum.txt", usecols=(1))

        # Find the k-edge MABE
        aims_y_max = y_axis_aims.max()
        aims_y_max_arg = y_axis_aims.argmax()
        aims_be = round(x_axis_aims[aims_y_max_arg], 4)
        aims_be_line = [i for i in np.linspace(-0.6, aims_y_max, num=len(y_axis_aims))]

        print("\nFHI-aims mean average binding energy (MABE):", aims_be, "eV")

        # Plot everything
        plt.xlabel("Energy / eV")
        plt.ylabel("Intensity")

        plt.plot(
            np.full((len(aims_be_line)), aims_be),
            aims_be_line,
            c="grey",
            linestyle="--",
            label=f"MABE = {aims_be} eV",
        )

        # Find the range of the spectrum to plot
        plot_x = []
        plot_y = []

        glob_max_y = max(y_axis_aims)
        glob_min_y = glob_max_y * gmp

        for c, y in enumerate(y_axis_aims):
            if y > glob_min_y:
                plot_x.append(x_axis_aims[c])
                plot_y.append(y)

        # Calculate the min and max ranges
        x_max = math.floor(max(plot_x))
        x_min = math.ceil(min(plot_x))
        y_max = max(plot_y)

        # Get the type of molecule
        # Enable for both basis and projector file structures
        try:
            with open(f"{run_loc}/{targ_at}{at_spec}/geometry.in", "r") as hole_geom:
                lines = hole_geom.readlines()

        except FileNotFoundError:
            with open(
                f"{run_loc}/{targ_at}{at_spec}/hole/geometry.in", "r"
            ) as hole_geom:
                lines = hole_geom.readlines()

        molecule = lines[4].split()[-1]

        # Plot the individual binding energies
        first_dirac = True
        for peak in xps:
            # Include the peak in the legend if first call
            if first_dirac is True:
                plt.axvline(
                    x=peak,
                    c="#9467bd",
                    ymax=0.25,
                    label="Individual binding energies",
                )
                first_dirac = False
            else:
                plt.axvline(x=peak, c="#9467bd", ymax=0.25)

        # Plot the spectrum
        plt.plot(plot_x, plot_y, label="Simulated XPS spectrum")
        plt.ylim((0, y_max + 1))
        plt.xlim(x_max, x_min)  # Reverse to match experimental XPS conventions
        plt.xticks(np.arange(x_min, x_max, 1))
        plt.legend(loc="upper right")

        if molecule != "None":
            plt.title(f"XPS spectrum of {molecule}")

        plt.savefig(f"{run_loc}/xps_spectrum.pdf")
        plt.savefig(f"{run_loc}/xps_spectrum.png")
        plt.show()



import math

import numpy as np


def _schmid_pseudo_voigt(domain, A, m, E, omega, asymmetry, a, b):
    """
    Apply broadening scheme for XPS spectra
    https://analyticalsciencejournals.onlinelibrary.wiley.com/doi/10.1002/sia.5521

    domain = linspace of x range and bin width
    A = intensity
    m = Gaussian-Lorentzian mixing parameter
    E = line centre (aka dirac peak)
    omega = full width at half maximum (omega > 0)
    asymmetry = True or False
    a = asymmetry parameter
    b = asymmetry translation parameter
    """

    if asymmetry is False:
        return A * (1 - m) * np.sqrt((4 * np.log(2)) / (math.pi * omega**2)) * np.exp(
            -(4 * np.log(2) / omega**2) * (domain - E) ** 2
        ) + A * m * (1 / (2 * np.pi)) * (omega / ((omega / 2) ** 2 + (domain - E) ** 2))

    else:
        omega_as = 2 * omega / (1 + np.exp(-a * domain - b))

        return A * (1 - m) * np.sqrt(
            (4 * np.log(2))
            / (np.pi * ((2 * omega_as) / (1 + np.exp(-a * ((domain - E) - b)))) ** 2)
        ) * np.exp(
            -(
                4
                * np.log(2)
                / (2 * omega_as / (1 + np.exp(-a * ((domain - E) - b)))) ** 2
            )
            * (domain - E) ** 2
        ) + A * m * (
            1 / (2 * np.pi)
        ) * (
            (2 * omega_as / (1 + np.exp(-a * ((domain - E) - b))))
            / (
                (((2 * omega_as) / (1 + np.exp(-a * ((domain - E) - b)))) / 2) ** 2
                + (domain - E) ** 2
            )
        )


def broaden(start, stop, A, m, dirac_peaks, omega, asymmetry, a, b):
    """
    Broaden dirac delta peaks

    start = beginning of x range
    stop = end of x range
    dirac_peaks = list of dirac delta peaks
    eta = full width at half maximum (eta > 0)
    """
    domain = np.linspace(start, stop, 100000)
    data = np.zeros([len(domain)])
    for i in dirac_peaks:
        V = _schmid_pseudo_voigt(domain, A, m, i, omega, asymmetry, a, b)
        data += V

    return data, domain
